from typing import Callable, DefaultDict, Optional, Tuple

import cv2 as cv
import numpy as np

from creevey.constants import PathOrStr


def record_mean_brightness(
    image: np.array, inpath: PathOrStr, log_dict: DefaultDict[str, dict]
) -> np.array:
    """
    Calculate mean image brightness

    Image is assumed to be grayscale if it has a single channel, RGB if
    it has three channels, RGBA if it has four. Brightness is calculated
    by converting to grayscale if necessary and then taking the mean
    pixel value.

    Parameters
    ----------
    image
    inpath
        Image input path
    log_dict
        Dictionary of image metadata

    Side effect
    -----------
    Adds a "mean_brightness" items to log_dict[inpath]
    """
    if len(image.shape) == 3:
        num_bands = image.shape[2]
    elif len(image.shape) == 2:
        num_bands = 1
    else:
        raise ValueError('Image array must have two or three dimensions')

    if num_bands == 1:
        image_gray = image
    elif num_bands == 3:
        image_gray = cv.cvtColor(src=image, code=cv.COLOR_RGB2GRAY)
    elif num_bands == 4:
        image_gray = cv.cvtColor(src=image, code=cv.COLOR_RGBA2GRAY)
    else:
        raise ValueError(
            f'{inpath} image has {num_bands} channels. Only 1-channel '
            f'grayscale, 3-channel RGB, and 4-channel RGBA images are '
            f'supported.'
        )
    log_dict[inpath]['mean_brightness'] = image_gray.mean()

    return image


def resize(
    image: np.array,
    shape: Optional[Tuple[int, int]] = None,
    min_dim: Optional[int] = None,
    **kwargs,
) -> np.array:
    """
    Resize input image

    `shape` or `min_dim` needs to be specified with `partial` before
    this function can be used in a Creevey pipeline.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    image
        NumPy array with two spatial dimensions and optionally an
        additional channel dimension
    shape
        Desired output shape in pixels in the form (height, width)
    min_dim
        Desired minimum spatial dimension in pixels; image will be
        resized so that it has this length along its smaller spatial
        dimension while preseving aspect ratio as closely as possible.
        Exactly one of `shape` and `min_dim` must be `None`.

    Returns
    -------
    NumPy array with specified shape
    """
    _validate_resize_inputs(shape, min_dim)
    if min_dim is not None:
        shape = _find_min_dim_shape(image, min_dim)
    resized = cv.resize(image, dsize=shape[::-1])
    return resized


def _validate_resize_inputs(shape, min_dim) -> None:
    if (shape is None) + (min_dim is None) == 1:
        pass
    else:
        raise ValueError('Exactly one of `shape` and `min_dim` must be None')


def _find_min_dim_shape(image, min_dim):
    in_height, in_width = image.shape[:2]
    aspect_ratio = in_width / in_height
    format = 'tall' if aspect_ratio < 1 else 'wide'
    if format == 'tall':
        out_width = min_dim
        out_height = round(out_width / aspect_ratio, 1)
    else:
        out_height = min_dim
        out_width = round(out_height * aspect_ratio, 1)
    return (int(out_height), int(out_width))


def centercrop(image: np.array, reduction_factor: float, **kwargs) -> np.array:
    """
    Crop the center out of an image

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    image
        Numpy array of an image. Function will handle 2D greyscale
        images, RGB, and RGBA image arrays
    reduction_factor
        scale of center cropped box, 1.0 would be the full image
        value of .4 means a box of .4*width and .4*height

    Returns
    -------
    Slice of input image corresponding to a cropped area around the center
    """
    height, width, *channels = image.shape

    w_scale = width * reduction_factor
    h_scale = height * reduction_factor

    left = int((width - w_scale) // 2)
    top = int((height - h_scale) // 2)
    right = int((width + w_scale) // 2)
    bottom = int((height + h_scale) // 2)

    return image[top:bottom, left:right]


def trim_padding(
    image: np.array, comparison_op: Callable, thresh: int, **kwargs
) -> np.array:
    """
    Remove padding from an image

    Remove rows and columns on the edges of the input image where the
    brightness on a scale of 0 to 1 satisfies `comparison_op` with
    respect to `thresh`. Brightness is evaluated by converting to
    grayscale and normalizing if necessary. For instance, using
    `thresh=.95` and `comparison_op=operator.gt` will result in removing
    near-white padding, while using using `thresh=.05` and
    `comparison_op=operator.lt` will remove near-black padding.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Assumptions
    -----------
    - Image is grayscale, RGB, or RGBA.
    - Pixel values are scaled between either 0 and 1 or 0 and 255. If
    image is scaled between 0 and 255, then some pixel has a value
    greater than 1.

    Parameters
    ----------
    image
        Numpy array of an image.
    comparison_op
        How to compare pixel values to `thresh`
    thresh
        Value to compare pixel values against

    Returns
    -------
    Slice of input image corresponding to a cropped area with padding
    removed
    """
    im_gray = _convert_to_grayscale(image)
    im_gray = normalize_pixel_values(im_gray)
    keep = ~comparison_op(im_gray, thresh)
    x, y, w, h = cv.boundingRect(cv.findNonZero(keep.astype(int)))
    return image[y : y + h, x : x + w]


def normalize_pixel_values(image: np.array) -> np.array:
    """
    Normalize image so that pixel values are between 0 and 1

    Assumes pixel values are scaled between either 0 and 1 or 0 and 255.
    """
    if image.max() > 1:
        return image / 255
    else:
        return image


def _convert_to_grayscale(image: np.array) -> np.array:
    """
    Convert image to grayscale.

    Assumes image is grayscale, RGB, or RGBA.
    """
    grayscale = image.ndim == 2
    if grayscale:
        im_gray = image
    else:
        rgba = image.shape[2] == 4
        if rgba:
            im_gray = cv.cvtColor(image, cv.COLOR_RGBA2GRAY)
        else:
            im_gray = cv.cvtColor(image, cv.COLOR_RGB2GRAY)
    return im_gray



from functools import partial
from pathlib import Path
from shutil import rmtree

import pytest

from creevey.path_funcs import join_outdir_filename_extension


TEST_DIR = Path(__file__).parent
SAMPLE_DATA_DIR = Path(TEST_DIR) / 'sample_data'
TEMP_DATA_DIR = SAMPLE_DATA_DIR / 'tmp'
IMAGE_FILENAMES = ['2RsJ8EQ', '2TqoToT', '2VocS58', '2scKPIp', '2TsO6Pc', '2SCv0q7']
IMAGE_URLS = [f'https://bit.ly/{filename}' for filename in IMAGE_FILENAMES]

keep_filename_save_png_in_tempdir = partial(
    join_outdir_filename_extension, outdir=TEMP_DATA_DIR, extension='.png'
)


@pytest.fixture
def generate_file_tree(scope='session'):
    if TEMP_DATA_DIR.exists():
        rmtree(str(TEMP_DATA_DIR))
    level0_dirs = ['tmp00', 'tmp01']
    level0_files = ['fake00.txt', 'fake01.png', 'fake02.JPG']
    for temp_dir in level0_dirs:
        dirpath = TEMP_DATA_DIR / temp_dir
        dirpath.mkdir(parents=True)
    for filename in level0_files:
        filepath = TEMP_DATA_DIR / filename
        filepath.touch()
    level2_dirpath = TEMP_DATA_DIR / level0_dirs[0] / 'temp10' / 'temp20'
    level2_dirpath.mkdir(parents=True)
    level3_files = ['fake30.pdf', 'fake31.BMP', 'fake32.txt']
    for filename in level3_files:
        filepath = level2_dirpath / filename
        filepath.touch()
    yield
    rmtree(str(TEMP_DATA_DIR))


def delete_file_if_exists(path: Path):
    if path.is_file():
        path.unlink()



from collections import defaultdict
import logging
from pathlib import Path
import time
from typing import Any, Callable, DefaultDict, Iterable, Optional, Tuple, Union

from joblib import delayed, Parallel
from numpy import iterable
import pandas as pd
from tqdm import tqdm

from creevey.constants import PathOrStr


class Pipeline:
    """
    Class for defining file processing pipelines.

    See Creevey's README for an example.

    Attributes
    ----------
    load_func
        Callable that takes a string or `Path` object as single
        positional argument, reads from the corresponding location, and
        returns some representation of its contents.
    ops
        Iterable of callables each of which takes a single positional
        argument. The first element of `ops` must accept the output of
        `load_func`, and each subsequent element must accept the output
        of the immediately preceding element. It is recommended that
        every element of `ops` take and return one common data structure
        (e.g. NumPy arrays for image data) so that those elements can
        be recombined easily.
    write_func
        Callable that takes the output of the last element of `ops` (or
        the output of `load_func` if `ops` is `None` or empty) and a
        string or `Path` object and writes the former to the location
        specified by the latter.
    """

    def __init__(
        self,
        load_func: Callable[[PathOrStr], Any],
        ops: Optional[Union[Callable[[Any], Any], Iterable[Callable[[Any], Any]]]],
        write_func: Callable[[Any, PathOrStr], None],
    ) -> None:
        """
        Compose the provided functions, and store them as attributes.

        Store `load_func`, `ops`, and `write_func` as attributes with
        the corresponding names. Create an additional attribute
        `pipeline_func` that composes those functions for when `run`
        is called.

        See the `Pipeline` docstring for information about the form that
        `load_func`, `ops`, and `write_func` are expected to take.
        """
        self.load_func = load_func
        if callable(ops):
            self.ops = [ops]
        elif iterable(ops):
            self.ops = ops
        elif ops is None:
            self.ops = []
        else:
            raise TypeError(
                'ops must be callable, an iterable of ' 'callables, or `None`'
            )
        self.write_func = write_func

    def pipeline_func(
        self,
        inpath: PathOrStr,
        outpath_func: PathOrStr,
        skip_existing: bool,
        log_dict: DefaultDict[str, dict],
        exceptions_to_catch: Optional[Union[Tuple, Tuple[Exception]]] = None,
    ) -> None:
        """
        Process one file

        Use `self.load_func` to load the file at `inpath` into memory,
        pipe the result through the functions in `self.ops`, and use
        `self.write_func` to write it to `outpath_func(inpath)`.

        If `skip_existing` is `True`, check up front whether
        `outpath_func(inpath)` exists. If it does, skip the file.

        Catch `exceptions_to_catch` if they arise during file processing.

        Parameters
        ----------
        inpath
        outpath_func
        skip_existing
        log_dict
        exceptions_to_catch

        Side effects
        ------------
        Record results in a dict within `log_dict[inpath]`:
            - `outpath_func(inpath)` as "outpath"
            - 0/1 indicating whether existing file was skipped as
            "skipped_existing"
            - 0/1 indicating whether exception of a type specified in
            `exceptions_to_catch` was handled during processing as
            "exception_handled"
            - Timestamp indicating when processing finished as
            "time_finished"
        """
        outpath = outpath_func(inpath)
        skipped_existing = False
        exception_handled = False

        if skip_existing and Path(outpath).is_file():
            skipped_existing = True
            logging.debug(
                f'Skipping {inpath} because there is already a file at corresponding '
                f'output path {outpath}'
            )
        else:
            if exceptions_to_catch is None:
                self._run_pipeline_func(inpath, outpath, log_dict=log_dict)
            else:
                try:
                    self._run_pipeline_func(inpath, outpath, log_dict=log_dict)
                except exceptions_to_catch as e:
                    exception_handled = True
                    logging.error(e, inpath)

        inpath_logs = log_dict[inpath]
        inpath_logs['outpath'] = outpath
        inpath_logs['skipped_existing'] = int(skipped_existing)
        inpath_logs['exception_handled'] = int(exception_handled)
        inpath_logs['time_finished'] = time.time()

    def _run_pipeline_func(self, inpath, outpath, **kwargs):
        # `kwargs` included to handle unused `log_dict` so that
        # `pipeline_func` does not have to change in
        # `CustomReportingPipeline`
        stage = self.load_func(inpath)
        for op in self.ops:
            stage = op(stage)
        self.write_func(stage, outpath)

    def run(
        self,
        inpaths: Iterable[PathOrStr],
        path_func: Callable[[PathOrStr], PathOrStr],
        n_jobs: int,
        skip_existing: bool = True,
        exceptions_to_catch: Optional[Union[Exception, Tuple[Exception]]] = None,
    ) -> pd.DataFrame:
        """
        Run the pipeline.

        Across `n_jobs` threads, for each path in `inpaths`, if
        `skip_existing` is `True` and `path_func` of that path exists,
        do not do anything. Otherwise, use `load_func` to get the
        resource from that path, pipe its output through `ops`, and
        write out the result with `write_func`.

        Parameters
        ----------
        inpaths
            Iterable of string or Path objects pointing to resources to
            be processed and written out.
        path_func
            Function that takes each input path and returns the desired
            corresponding output path.
        n_jobs
            Number of threads to use.
        skip_existing
            Boolean indicating whether to skip items that would result
            in overwriting an existing file or to overwrite any such
            files.
        exceptions_to_catch
            Tuple of exception types to catch. An exception of one of
            these types will be logged with logging level ERROR and the
            relevant file will be skipped.

        Side effects
        ------------
        Logs a warning when `skip_existing` is `True`.

        Returns
        -------
        Pandas DataFrame "run report" with each input path as its
        index and columns indicating the corresponding output path (
        "outpath"), whether processing was skipped because a file
        already existed at the output path ("skipped_existing"),
        whether processing failed due to an exception in
        `exceptions_to_catch` ("exception_handled"), and a timestamp
        indicating when processing complete ("time_finished").
        """
        if skip_existing:
            logging.warning(
                'Skipping files where a file exists at the output '
                'location. Pass `skip_existing=False` to overwrite '
                'existing files instead.'
            )

        log_dict = defaultdict(dict)

        Parallel(n_jobs=n_jobs, prefer='threads')(
            delayed(self.pipeline_func)(
                path, path_func, skip_existing, log_dict, exceptions_to_catch
            )
            for path in tqdm(inpaths)
        )

        run_report = pd.DataFrame.from_dict(log_dict, orient='index')

        return run_report


class CustomReportingPipeline(Pipeline):
    """
    Class for defining file processing pipelines with custom run
    recording.

    Differences from Pipeline parent class:

    - `load_func`, each element of `ops`, and `write_func` must each
    accept the string or `Path` object indicating the input item's
    location as an additional positional argument.
    - Each element of `ops` and `write_func` must each accept a
    `defaultdict(dict)` object as an additional positional argument.
    Functions defined in Creevey call this item `log_dict`.

    Inside those functions, adding items to `log_dict[inpath]` causes
    them to be added to the "run record" DataFrame that `self.run`
    returns. See Creevey's README for further explanation.
    """

    def _run_pipeline_func(self, inpath, outpath, log_dict):
        stage = self.load_func(inpath, log_dict=log_dict)
        for op in self.ops:
            stage = op(stage, inpath=inpath, log_dict=log_dict)
        self.write_func(stage, outpath, inpath=inpath, log_dict=log_dict)



import cv2 as cv
import numpy as np

from creevey.constants import PathOrStr
from creevey.load_funcs import get_response


def load_image_from_url(inpath: str, **kwargs) -> np.array:
    """
    Download an image

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    inpath
        Image URL

    Returns
    -------
    Image as NumPy array
    """
    response = get_response(inpath)
    image = _load_image_from_response(response)
    return image


def _load_image_from_response(response):
    image = np.asarray(bytearray(response.content), dtype="uint8")
    image = cv.imdecode(image, cv.IMREAD_UNCHANGED)
    if image is None:
        raise ValueError(f'{response.url} failed to load')
    _convert_to_rgb_if_needed(image)

    return image


def load_image_from_disk(inpath: PathOrStr, **kwargs) -> np.array:
    """
    Load image from disk

    Assumes that image is RGB(A) if it has at least three channels.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    inpath
        Path to local image file

    Returns
    -------
    Image file contents as a NumPy array

    Raises
    ------
    ValueError if image fails to load
    """
    # As of mid-2018 OpenCV is faster than Matplotlib or Pillow for IO
    image = cv.imread(str(inpath), cv.IMREAD_UNCHANGED)
    if image is None:
        raise ValueError(f'{inpath} failed to load')
    _convert_to_rgb_if_needed(image)
    return image


def _convert_to_rgb_if_needed(image: np.array) -> np.array:
    """
    Reverse order of first three channels if image has at least three

    Used as a helper function when loading images with OpenCV, which
    uses BGR channel ordering rather than the more common RGB.
    """
    num_channels = 1 if len(image.shape) == 2 else image.shape[2]
    if num_channels >= 3:
        # OpenCV loads as BGR, so reverse order of first three channels
        image[:, :, :3] = image[:, :, 2::-1]
    return image



from pathlib import Path
from typing import Optional
import uuid

from creevey.constants import PathOrStr


def join_outdir_filename_extension(
    path: PathOrStr, outdir: PathOrStr, extension: Optional[str] = None
) -> Path:
    """
    Construct path by combining specified `outdir`, filename from
    `path`, and (optionally) specified extension

    Parameters
    ----------
    path
        Path with desired filename
    outdir
        Desired output directory
    extension
        Desired output file extension. If `None`, keep extension from
        `path`.
    """
    filename = Path(Path(path).name)
    if extension is not None:
        extension = extension if extension.startswith('.') else '.' + extension
        filename = filename.stem + extension
    outpath = Path(outdir) / filename
    return outpath


def join_outdir_hashed_path_extension(
    path: PathOrStr, outdir: PathOrStr, extension: Optional[str] = None
) -> Path:
    """
    Construct path by combining specified `outdir`, filename derived by
    hashing `path`, and (optionally) specified extension.

    Parameters
    ----------
    path
        Input path to be hashed to generate output filename
    outdir
        Desired output directory
    extension
        Desired output extension. If `None`, keep extension from `path`.
    """
    filename = str(uuid.uuid5(uuid.NAMESPACE_DNS, str(path)))
    filename = filename.replace('-', '')
    if extension is None:
        extension = path.suffix
    outpath = join_outdir_filename_extension(filename, outdir, extension)
    return outpath


def replace_dir(path: PathOrStr, outdir: PathOrStr) -> Path:
    return Path(outdir) / Path(path).name



from functools import partial

from creevey.util import find_files_with_extensions

IMAGE_EXTENSIONS = ["jpeg", "jpg", "png", "bmp", "gif"]


find_image_files = partial(find_files_with_extensions, extensions=IMAGE_EXTENSIONS)



from pathlib import Path

import cv2 as cv
import numpy as np

from creevey.constants import PathOrStr


def write_image(image: np.array, path: PathOrStr, **kwargs) -> None:
    """
    Write image to specified path

    Create output directory if it does not exist, with try/except for
    thread safety.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    image
        Image as a NumPy array
    path
        Desired output path
    """
    outdir = Path(path).parent
    if not outdir.exists():
        # need try/except in case directory does not exist but at this
        # point process switches to another thread that creates it
        # before coming back to this one
        try:
            outdir.mkdir(parents=True)
        except OSError:
            pass
    num_channels = 1 if len(image.shape) == 2 else image.shape[2]
    if num_channels >= 3:
        # OpenCV wants to write BGR, so reverse order of first three
        # channels
        image[:, :, :3] = image[:, :, 2::-1]
    cv.imwrite(str(path), image)



import os
from pathlib import Path
from typing import Iterable, List

from creevey.constants import PathOrStr


def find_files_with_extensions(
    search_dir: PathOrStr, extensions: Iterable[str]
) -> List[Path]:
    """
    Find files with one of the specified extensions.

    Extension matching is case-insensitive.

    Parameters
    ----------
    search_dir
        Directory to search
    img_extensions
        Extensions to search for. The initial "." can be included or
        not.

    Returns
    -------
    List of `Path` objects specifying locations of all files recursively
    within `search_dir` that have one of the extensions in `extensions`.
    """
    extensions = [item if item.startswith('.') else '.' + item for item in extensions]
    file_list = [
        Path(dirpath) / fn
        for dirpath, dirname, filenames in os.walk(str(search_dir))
        for fn in filenames
        if Path(fn).suffix.lower() in extensions
    ]
    return file_list



from pathlib import Path
from setuptools import find_packages, setup

version_file_path = Path(__file__).parent / 'creevey' / '_version.py'

with open(version_file_path, 'r') as version_file:
    exec(version_file.read())

with open('README.md') as r:
    readme = r.read()

regular_packages = [
    'boto3',
    'joblib',
    'opencv-contrib-python',
    'pandas',
    'pillow',
    'requests',
    'retrying',
    'scikit-learn',
    'tqdm',
]
dev_packages = [
    'black',
    'flake8',
    'flake8-docstrings',
    'flake8-import-order',
    'matplotlib',
    'pytest',
    'pytest-cov',
    'responses',
]


setup(
    name='creevey',
    version=__version__,
    description='Bulk image processing',
    long_description=readme,
    packages=find_packages(exclude=('tests')),
    install_requires=regular_packages,
    extras_require={'dev': regular_packages + dev_packages},
    author='Greg Gandenberger',
    author_email='gsganden@gmail.com',
    url='https://github.com/ShopRunner/creevey',
    download_url=f'https://github.com/ShopRunner/creevey/tarball/{__version__}',
    long_description_content_type='text/markdown',
)



from creevey import Pipeline
from creevey.load_funcs.image import load_image_from_url
from creevey.write_funcs.image import write_image

download_image_pipeline = Pipeline(
    load_func=load_image_from_url, ops=[], write_func=write_image
)



import logging
import threading

import requests
from retrying import retry

threadLocal = threading.local()


def _is_connection_error(exception):
    return isinstance(
        exception, (requests.exceptions.ConnectionError, _HTTPErrorToRetry)
    )


@retry(
    retry_on_exception=_is_connection_error,
    wait_exponential_multiplier=1000,
    wait_exponential_max=10000,
    stop_max_attempt_number=10,
)
def get_response(url: str, **kwargs) -> None:
    """
    Make a GET request to the specified URL and return the response.

    Maintain a common session within each thread to reduce request
    overhead.

    Error Handling
    --------------
    Retry up to ten times on `requests.exceptions.ConnectionError`
    instances and 5xx status codes, waiting 2^x seconds between each
    retry, with a max of 10 seconds.

    Log an error without retrying or raising for 403 and 404 status
    codes.

    Raise `requests.exceptions.HTTPError` for other non-200 status
    codes.

    `kwargs` is included only for compatibility with the
    `CustomReportingPipeline` class.

    Parameters
    ----------
    url
        URL of file to download
    """
    session = _get_session()
    response = session.get(url)
    _check_status(url, response)
    return response


def _get_session():
    if getattr(threadLocal, 'session', None) is None:
        threadLocal.session = requests.Session()
    return threadLocal.session


def _check_status(url, response):
    if response.status_code == 200:
        pass
    elif 600 > response.status_code > 499:
        logging.debug(f'Retrying {url} with status code {response.status_code}')
        raise _HTTPErrorToRetry
    elif response.status_code in [403, 404]:
        logging.error(
            f'Failed to download {url} with status code {response.status_code}'
        )
    else:
        response.raise_for_status()


class _HTTPErrorToRetry(requests.exceptions.HTTPError):
    pass
